# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CSPReport',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('json', models.TextField()),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
    ]



"""Settings for cspreports tests."""
from __future__ import unicode_literals

SECRET_KEY = 'CSP_REPORTS_TESTS'

INSTALLED_APPS = [
    'cspreports',
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}



from __future__ import unicode_literals

from datetime import datetime

import mock
from django.http import HttpRequest
from django.test import RequestFactory, SimpleTestCase, TestCase
from django.test.utils import override_settings
from django.utils import timezone
from mock import patch

from cspreports import utils
from cspreports.models import CSPReport
from cspreports.utils import get_midnight, parse_date_input

JSON_CONTENT_TYPE = 'application/json'


class UtilsTest(TestCase):

    def test_config(self):
        """ Test that the various CSP_REPORTS_X settings correctly control which handlers are
            called.
        """
        mock_paths = [
            "cspreports.utils.email_admins",
            "cspreports.utils.save_report",
            "cspreports.utils.log_report",
        ]
        corresponding_settings = [
            "CSP_REPORTS_EMAIL_ADMINS",
            "CSP_REPORTS_SAVE",
            "CSP_REPORTS_LOG",
        ]
        for i in range(len(mock_paths)):
            mocks = [mock.patch(path) for path in mock_paths]
            settings_overrides = {
                setting: True if j == i else False
                for j, setting in enumerate(corresponding_settings)
            }
            with override_settings(**settings_overrides):
                with mocks[0] as mocked_object_0, mocks[1] as mocked_object_1, mocks[2] as mocked_object_2:
                    mocked_objects = [mocked_object_0, mocked_object_1, mocked_object_2]
                    request = HttpRequest()
                    utils.process_report(request)
                    for k, mocked_object in enumerate(mocked_objects):
                        if k == i:
                            self.assertTrue(mocked_object.called)
                        else:
                            self.assertFalse(mocked_object.called)

    def test_save_report(self):
        """ Test that the `save_report` handler correctly saves to the DB. """
        assert CSPReport.objects.count() == 0  # sanity
        body = '{"document-uri": "http://example.com/"}'
        request = RequestFactory(HTTP_USER_AGENT='Agent007').post('/dummy/', body, content_type=JSON_CONTENT_TYPE)

        utils.save_report(request)

        reports = CSPReport.objects.all()
        self.assertQuerysetEqual(reports.values_list('user_agent'), [('Agent007', )], transform=tuple)
        self.assertEqual(reports[0].json, body)

    def test_save_report_no_agent(self):
        """Test that the `save_report` handler correctly handles missing user agent header."""
        request = RequestFactory().post('/dummy/', '{"document-uri": "http://example.com/"}',
                                        content_type=JSON_CONTENT_TYPE)

        utils.save_report(request)

        self.assertQuerysetEqual(CSPReport.objects.values_list('user_agent'), [('', )], transform=tuple)

    @override_settings(CSP_REPORTS_LOG_LEVEL='warning')
    def test_log_report(self):
        """ Test that the `log_report` handler correctly logs at the right level. """
        request = HttpRequest()
        report = '{"document-uri": "http://example.com/"}'
        formatted_report = utils.format_report(report)
        request._body = report
        with mock.patch("cspreports.utils.logger.warning") as warning_mock:
            utils.log_report(request)
            self.assertTrue(warning_mock.called)
            log_message = warning_mock.call_args[0][0] % warning_mock.call_args[0][1:]
            self.assertTrue(formatted_report in log_message)

    def test_email_admins(self):
        """ Test that the `email_admins` handler correctly sends an email. """
        request = HttpRequest()
        report = '{"document-uri": "http://example.com/"}'
        formatted_report = utils.format_report(report)
        request._body = report
        # Note that we are mocking the *Django* mail_admins function here.
        with mock.patch("cspreports.utils.mail_admins") as mock_mail_admins:
            utils.email_admins(request)
            self.assertTrue(mock_mail_admins.called)
            message = mock_mail_admins.call_args[0][1]
            self.assertTrue(formatted_report in message)

    def test_format_report_handles_invalid_json(self):
        """ Test that `format_report` doesn't trip up on invalid JSON.
            Note: this is about not getting a ValueError, rather than any kind of security thing.
        """
        invalid_json = '{"key": undefined_variable, nonsense here}'
        try:
            formatted = utils.format_report(invalid_json)
        except ValueError as e:
            self.fail("format_report did not handle invalid JSON: %s" % e)
        # we expect our invalid JSON to remain in the output, as is
        self.assertTrue(invalid_json in formatted)

    def test_run_additional_handlers(self):
        """ Test that the run_additional_handlers function correctly calls each of the specified custom
            handler functions.
        """
        # utils stores a cache of the handlers (for efficiency, so kill that)
        utils._additional_handlers = None
        request = HttpRequest()
        with override_settings(
            CSP_REPORTS_ADDITIONAL_HANDLERS=["cspreports.tests.test_utils.my_handler"],
            CSP_REPORTS_EMAIL_ADMINS=False,
            CSP_REPORTS_LOG=False,
            CSP_REPORTS_SAVE=False,
        ):
            utils.process_report(request)
            self.assertTrue(request.my_handler_called)


def my_handler(request):
    # just set an attribute so that we can see that this function has been called
    request.my_handler_called = True


class TestParseDateInput(SimpleTestCase):
    """Test `parse_date_input` function."""

    def test_aware(self):
        with self.settings(USE_TZ=True, TIME_ZONE='Europe/Prague'):
            self.assertEqual(parse_date_input('2016-05-25'), timezone.make_aware(datetime(2016, 5, 25)))

    def test_naive(self):
        with self.settings(USE_TZ=False):
            self.assertEqual(parse_date_input('2016-05-25'), datetime(2016, 5, 25))

    def test_invalid_date(self):
        with self.assertRaisesMessage(ValueError, 'is not a valid date.'):
            parse_date_input('2016-13-25')

    def test_invalid_input(self):
        with self.assertRaisesMessage(ValueError, 'is not a valid date.'):
            parse_date_input('INVALID')


class TestGetMidnight(SimpleTestCase):
    """Test `get_midnight` function."""

    def test_aware(self):
        with self.settings(USE_TZ=True, TIME_ZONE='Europe/Prague'):
            # 00:05 in CEST is 22:05 day before in UTC
            mock_now = datetime(2016, 4, 26, 22, 5, tzinfo=timezone.utc)
            with patch('cspreports.utils.now', return_value=mock_now):
                self.assertEqual(get_midnight(), datetime(2016, 4, 26, 22, 0, tzinfo=timezone.utc))

    def test_naive(self):
        with self.settings(USE_TZ=False):
            mock_now = datetime(2016, 4, 27, 12, 34)
            with patch('cspreports.utils.now', return_value=mock_now):
                self.assertEqual(get_midnight(), datetime(2016, 4, 27))



"""Test `summary` module."""
from __future__ import unicode_literals

from datetime import datetime

from django.test import SimpleTestCase, TestCase
from mock import sentinel
from six import assertCountEqual

from cspreports.models import CSPReport
from cspreports.summary import CspReportSummary, ViolationInfo, collect, get_root_uri

from .utils import create_csp_report


class TestGetRootUri(SimpleTestCase):
    """Test `get_root_uri` function."""

    def test_root_uri(self):
        self.assertEqual(get_root_uri(''), '')
        self.assertEqual(get_root_uri('self'), 'self')
        self.assertEqual(get_root_uri('http://example.cz/'), 'http://example.cz/')
        self.assertEqual(get_root_uri('http://example.cz/path/'), 'http://example.cz/path/')
        self.assertEqual(get_root_uri('http://example.cz/path/?query=value'), 'http://example.cz/path/')
        self.assertEqual(get_root_uri('http://example.cz/path/#fragment'), 'http://example.cz/path/')


class TestViolationInfo(SimpleTestCase):
    """Test `ViolationInfo` class."""

    def test_empty(self):
        info = ViolationInfo(sentinel.root_uri)

        self.assertEqual(info.root_uri, sentinel.root_uri)
        self.assertEqual(info.count, 0)
        self.assertEqual(info.examples, [])

    def test_append(self):
        info = ViolationInfo(sentinel.root_uri)
        info.append(sentinel.report)

        self.assertEqual(info.count, 1)
        self.assertEqual(info.examples, [sentinel.report])

    def test_append_top(self):
        # Test appending over the limit
        info = ViolationInfo(sentinel.root_uri, top=2)
        info.append(sentinel.report_1)
        info.append(sentinel.report_2)

        info.append(sentinel.report_over)

        self.assertEqual(info.count, 3)
        self.assertEqual(info.examples, [sentinel.report_1, sentinel.report_2])


class TestCspReportSummary(SimpleTestCase):
    """Test `CspReportSummary` class."""

    def test_render_empty(self):
        summary = CspReportSummary(sentinel.since, sentinel.to)

        output = summary.render()

        self.assertIn('CSP report summary', output)

    def test_render_reports(self):
        summary = CspReportSummary(sentinel.since, sentinel.to)
        summary.total_count = 42
        summary.valid_count = 32
        summary.invalid_count = 10
        violation = ViolationInfo('http://example.cz/')
        violation.append(CSPReport())
        summary.sources = [violation]
        summary.blocks = [violation]

        output = summary.render()

        self.assertIn('CSP report summary', output)
        self.assertIn('Violation sources', output)
        self.assertIn('Blocked URIs', output)


class TestCollect(TestCase):
    """Test `collect` function."""

    def test_no_reports(self):
        summary = collect(datetime(1970, 1, 1), datetime(1970, 12, 31))

        self.assertEqual(summary.total_count, 0)
        self.assertEqual(summary.valid_count, 0)
        assertCountEqual(self, summary.sources, ())
        assertCountEqual(self, summary.blocks, ())
        self.assertEqual(summary.invalid_count, 0)
        assertCountEqual(self, summary.invalid_reports, ())

    def test_invalid_reports(self):
        report1 = create_csp_report(datetime(1970, 1, 1, 12))
        report2 = create_csp_report(datetime(1970, 1, 1, 12))
        summary = collect(datetime(1970, 1, 1), datetime(1970, 12, 31))

        self.assertEqual(summary.total_count, 2)
        self.assertEqual(summary.valid_count, 0)
        assertCountEqual(self, summary.sources, ())
        assertCountEqual(self, summary.blocks, ())
        self.assertEqual(summary.invalid_count, 2)
        assertCountEqual(self, summary.invalid_reports, (report1, report2))

    def test_valid_reports(self):
        report1 = create_csp_report(datetime(1970, 1, 1, 12), is_valid=True, document_uri='http://example.cz/',
                                    blocked_uri='http://example.evil/')
        report2 = create_csp_report(datetime(1970, 1, 1, 12), is_valid=True,
                                    document_uri='http://example.cz/?key=value',
                                    blocked_uri='http://example.evil/')
        summary = collect(datetime(1970, 1, 1), datetime(1970, 12, 31))

        self.assertEqual(summary.total_count, 2)
        self.assertEqual(summary.valid_count, 2)
        # Check sources
        self.assertEqual(len(summary.sources), 1)
        source = summary.sources[0]
        self.assertEqual(source.root_uri, 'http://example.cz/')
        self.assertEqual(source.count, 2)
        assertCountEqual(self, source.examples, (report1, report2))
        # Check blocks
        self.assertEqual(len(summary.blocks), 1)
        block = summary.blocks[0]
        self.assertEqual(block.root_uri, 'http://example.evil/')
        self.assertEqual(block.count, 2)
        assertCountEqual(self, block.examples, (report1, report2))
        # Check invalid
        self.assertEqual(summary.invalid_count, 0)
        assertCountEqual(self, summary.invalid_reports, ())



from __future__ import unicode_literals

import json
import logging
from datetime import datetime
from importlib import import_module

from django.conf import settings
from django.core.mail import mail_admins
from django.utils.dateparse import parse_date
from django.utils.timezone import localtime, make_aware, now

from cspreports.models import CSPReport

logger = logging.getLogger(getattr(settings, "CSP_REPORTS_LOGGER_NAME", "CSP Reports"))


def process_report(request):
    """ Given the HTTP request of a CSP violation report, log it in the required ways. """
    if config.EMAIL_ADMINS:
        email_admins(request)
    if config.LOG:
        log_report(request)
    if config.SAVE:
        save_report(request)
    if config.ADDITIONAL_HANDLERS:
        run_additional_handlers(request)


def format_report(jsn):
    """ Given a JSON report, return a nicely formatted (i.e. with indentation) string.
        This should handle invalid JSON (as the JSON comes from the browser/user).
        We trust that Python's json library is secure, but if the JSON is invalid then we still
        want to be able to display it, rather than tripping up on a ValueError.
    """
    if isinstance(jsn, bytes):
        jsn = jsn.decode('utf-8')
    try:
        return json.dumps(json.loads(jsn), indent=4, sort_keys=True, separators=(',', ': '))
    except ValueError:
        return "Invalid JSON. Raw dump is below.\n\n" + jsn


def email_admins(request):
    user_agent = request.META.get('HTTP_USER_AGENT', '')
    report = format_report(request.body)
    message = "User agent:\n%s\n\nReport:\n%s" % (user_agent, report)
    mail_admins("CSP Violation Report", message)


def log_report(request):
    func = getattr(logger, config.LOG_LEVEL)
    func("Content Security Policy violation: %s", format_report(request.body))


def save_report(request):
    report = CSPReport.from_message(request.body.decode(request.encoding or settings.DEFAULT_CHARSET))
    report.user_agent = request.META.get('HTTP_USER_AGENT', '')
    report.save()


def run_additional_handlers(request):
    for handler in get_additional_handlers():
        handler(request)


class Config(object):
    """ Configuration with defaults, each of which is overrideable in django settings. """

    # Defaults, these are overridden using "CSP_REPORTS_"-prefixed versions in settings.py
    EMAIL_ADMINS = True
    LOG = True
    LOG_LEVEL = 'warning'
    SAVE = True
    ADDITIONAL_HANDLERS = []

    def __getattribute__(self, name):
        try:
            return getattr(settings, "%s%s" % ("CSP_REPORTS_", name))
        except AttributeError:
            return super(Config, self).__getattribute__(name)


config = Config()
_additional_handlers = None


def get_additional_handlers():
    """ Returns the actual functions from the dotted paths specified in ADDITIONAL_HANDLERS. """
    global _additional_handlers
    if not isinstance(_additional_handlers, list):
        handlers = []
        for name in config.ADDITIONAL_HANDLERS:
            module_name, function_name = name.rsplit('.', 1)
            function = getattr(import_module(module_name), function_name)
            handlers.append(function)
        _additional_handlers = handlers
    return _additional_handlers


def parse_date_input(value):
    """Return datetime based on the user's input.

    @param value: User's input
    @type value: str
    @raise ValueError: If the input is not valid.
    @return: Datetime of the beginning of the user's date.
    """
    try:
        limit = parse_date(value)
    except ValueError:
        limit = None
    if limit is None:
        raise ValueError("'{}' is not a valid date.".format(value))
    limit = datetime(limit.year, limit.month, limit.day)
    if settings.USE_TZ:
        limit = make_aware(limit)
    return limit


def get_midnight():
    """Return last midnight in localtime as datetime.

    @return: Midnight datetime
    """
    limit = now()
    if settings.USE_TZ:
        limit = localtime(limit)
    return limit.replace(hour=0, minute=0, second=0, microsecond=0)



"""Test for `cspreports.models`."""
from __future__ import unicode_literals

import json

from django.test import SimpleTestCase
from django.utils.six import text_type

from cspreports.models import CSPReport


class TestCSPReport(SimpleTestCase):
    """Basic tests of CSP report model."""

    def test_nice_report_empty(self):
        self.assertEqual(CSPReport(json=None).nice_report, '[no CSP report data]')
        self.assertEqual(CSPReport(json='').nice_report, '[no CSP report data]')

    def test_nice_report_invalid_json(self):
        self.assertEqual(CSPReport(json='Not a JSON').nice_report, "Invalid CSP report: 'Not a JSON'")

    def test_nice_report_invalid_report(self):
        self.assertEqual(CSPReport(json=json.dumps({})).nice_report, "Invalid CSP report: {}")
        self.assertEqual(CSPReport(json=json.dumps({'key': 'value'})).nice_report,
                         'Invalid CSP report: {\n    "key": "value"\n}')

    def test_nice_report(self):
        self.assertJSONEqual(CSPReport(json=json.dumps({'csp-report': {}})).nice_report, {})
        self.assertJSONEqual(CSPReport(json=json.dumps({'csp-report': {'key': 'value'}})).nice_report, {'key': 'value'})

    def test_text_representation(self):
        self.assertEqual(text_type(CSPReport(json='')), '[no CSP report data]')
        self.assertEqual(text_type(CSPReport(json='Not a JSON')), "Invalid CSP report: 'Not a JSON'")
        self.assertJSONEqual(text_type(CSPReport(json=json.dumps({'csp-report': {}}))), {})
        self.assertJSONEqual(text_type(CSPReport(json=json.dumps({'csp-report': {'key': 'value'}}))), {'key': 'value'})


class TestFromMessage(SimpleTestCase):
    """Test `CSPReport.from_message` method."""

    def test_invalid_json(self):
        # Test report which is not a valid JSON.
        report = CSPReport.from_message('NOT_A_JSON')

        self.assertFalse(report.is_valid)
        self.assertEqual(report.json, 'NOT_A_JSON')
        self.assertIsNone(report.document_uri)

    def test_invalid_report(self):
        # Test report which is only a valid JSON.
        report = CSPReport.from_message('{}')

        self.assertFalse(report.is_valid)
        self.assertEqual(report.json, '{}')
        self.assertIsNone(report.document_uri)

    def test_empty_csp_report(self):
        # Test JSON with empty 'csp-report' object.
        report = CSPReport.from_message("{'csp-report': {}}")

        self.assertFalse(report.is_valid)
        self.assertEqual(report.json, "{'csp-report': {}}")
        self.assertIsNone(report.document_uri)

    def test_partial_csp_report(self):
        # Test partial, but invalid CSP report
        data = {'csp-report': {'blocked-uri': 'self', 'violated-directive': 'inline script base restriction'}}
        message = json.dumps(data)
        report = CSPReport.from_message(message)

        self.assertFalse(report.is_valid)
        self.assertEqual(report.json, message)
        self.assertIsNone(report.document_uri)
        self.assertEqual(report.blocked_uri, 'self')
        self.assertEqual(report.violated_directive, 'inline script base restriction')

    def test_valid_csp_1(self):
        # Test valid CSP report according to CSP 1.0
        data = {'csp-report': {'document-uri': 'http://protected.example.cz/',
                               'referrer': 'http://referrer.example.cz/',
                               'blocked-uri': 'http://dangerous.example.cz/',
                               'violated-directive': 'Very protective directive.',
                               'original-policy': 'Nothing is allowed.'}}
        message = json.dumps(data)
        report = CSPReport.from_message(message)

        self.assertTrue(report.is_valid)
        self.assertEqual(report.json, message)
        self.assertEqual(report.document_uri, 'http://protected.example.cz/')
        self.assertEqual(report.referrer, 'http://referrer.example.cz/')
        self.assertEqual(report.blocked_uri, 'http://dangerous.example.cz/')
        self.assertEqual(report.violated_directive, 'Very protective directive.')
        self.assertEqual(report.original_policy, 'Nothing is allowed.')
        self.assertIsNone(report.effective_directive)

    def test_valid_csp_2_plus(self):
        # Test valid CSP report according to CSP level >= 2.0
        data = {'csp-report': {'document-uri': 'http://protected.example.cz/',
                               'referrer': 'http://referrer.example.cz/',
                               'blocked-uri': 'http://dangerous.example.cz/',
                               'violated-directive': 'Very protective directive.',
                               'original-policy': 'Nothing is allowed.',
                               'source-file': 'nasty-script.js'}}
        message = json.dumps(data)
        report = CSPReport.from_message(message)

        self.assertTrue(report.is_valid)
        self.assertEqual(report.json, message)
        self.assertEqual(report.document_uri, 'http://protected.example.cz/')
        self.assertEqual(report.referrer, 'http://referrer.example.cz/')
        self.assertEqual(report.blocked_uri, 'http://dangerous.example.cz/')
        self.assertEqual(report.violated_directive, 'Very protective directive.')
        self.assertEqual(report.original_policy, 'Nothing is allowed.')
        self.assertEqual(report.source_file, 'nasty-script.js')
        self.assertIsNone(report.effective_directive)

    def test_valid_empty_fields(self):
        # Test valid CSP report according to CSP 1.0 with some fields with empty values
        data = {'csp-report': {'document-uri': 'http://protected.example.cz/',
                               'referrer': '',
                               'blocked-uri': '',
                               'violated-directive': 'Very protective directive.',
                               'original-policy': 'Nothing is allowed.'}}
        message = json.dumps(data)
        report = CSPReport.from_message(message)

        self.assertTrue(report.is_valid)
        self.assertEqual(report.json, message)
        self.assertEqual(report.document_uri, 'http://protected.example.cz/')
        self.assertEqual(report.referrer, '')
        self.assertEqual(report.blocked_uri, '')
        self.assertEqual(report.violated_directive, 'Very protective directive.')
        self.assertEqual(report.original_policy, 'Nothing is allowed.')
        self.assertIsNone(report.effective_directive)

    def test_valid_line_number(self):
        # Test valid line number is extracted.
        data = {'csp-report': {'document-uri': 'http://protected.example.cz/',
                               'referrer': 'http://referrer.example.cz/',
                               'blocked-uri': 'http://dangerous.example.cz/',
                               'violated-directive': 'Very protective directive.',
                               'original-policy': 'Nothing is allowed.',
                               'line-number': 666}}
        message = json.dumps(data)
        report = CSPReport.from_message(message)

        self.assertTrue(report.is_valid)
        self.assertEqual(report.json, message)
        self.assertEqual(report.document_uri, 'http://protected.example.cz/')
        self.assertEqual(report.referrer, 'http://referrer.example.cz/')
        self.assertEqual(report.blocked_uri, 'http://dangerous.example.cz/')
        self.assertEqual(report.violated_directive, 'Very protective directive.')
        self.assertEqual(report.original_policy, 'Nothing is allowed.')
        self.assertEqual(report.line_number, 666)

    def test_invalid_line_number(self):
        # Test invalid line number is ignored.
        data = {'csp-report': {'document-uri': 'http://protected.example.cz/',
                               'referrer': 'http://referrer.example.cz/',
                               'blocked-uri': 'http://dangerous.example.cz/',
                               'violated-directive': 'Very protective directive.',
                               'original-policy': 'Nothing is allowed.',
                               'line-number': -666}}
        message = json.dumps(data)
        report = CSPReport.from_message(message)

        self.assertTrue(report.is_valid)
        self.assertEqual(report.json, message)
        self.assertEqual(report.document_uri, 'http://protected.example.cz/')
        self.assertEqual(report.referrer, 'http://referrer.example.cz/')
        self.assertEqual(report.blocked_uri, 'http://dangerous.example.cz/')
        self.assertEqual(report.violated_directive, 'Very protective directive.')
        self.assertEqual(report.original_policy, 'Nothing is allowed.')
        self.assertIsNone(report.line_number)

    def test_valid_disposition(self):
        # Test valid disposition is extracted.
        data = {'csp-report': {'document-uri': 'http://protected.example.cz/',
                               'referrer': 'http://referrer.example.cz/',
                               'blocked-uri': 'http://dangerous.example.cz/',
                               'violated-directive': 'Very protective directive.',
                               'original-policy': 'Nothing is allowed.',
                               'disposition': 'report'}}
        message = json.dumps(data)
        report = CSPReport.from_message(message)

        self.assertTrue(report.is_valid)
        self.assertEqual(report.json, message)
        self.assertEqual(report.document_uri, 'http://protected.example.cz/')
        self.assertEqual(report.referrer, 'http://referrer.example.cz/')
        self.assertEqual(report.blocked_uri, 'http://dangerous.example.cz/')
        self.assertEqual(report.violated_directive, 'Very protective directive.')
        self.assertEqual(report.original_policy, 'Nothing is allowed.')
        self.assertEqual(report.disposition, 'report')

    def test_invalid_disposition(self):
        # Test invalid disposition is ignored.
        data = {'csp-report': {'document-uri': 'http://protected.example.cz/',
                               'referrer': 'http://referrer.example.cz/',
                               'blocked-uri': 'http://dangerous.example.cz/',
                               'violated-directive': 'Very protective directive.',
                               'original-policy': 'Nothing is allowed.',
                               'disposition': 'INVALID'}}
        message = json.dumps(data)
        report = CSPReport.from_message(message)

        self.assertTrue(report.is_valid)
        self.assertEqual(report.json, message)
        self.assertEqual(report.document_uri, 'http://protected.example.cz/')
        self.assertEqual(report.referrer, 'http://referrer.example.cz/')
        self.assertEqual(report.blocked_uri, 'http://dangerous.example.cz/')
        self.assertEqual(report.violated_directive, 'Very protective directive.')
        self.assertEqual(report.original_policy, 'Nothing is allowed.')
        self.assertIsNone(report.disposition)



from __future__ import unicode_literals

from django.conf.urls import url
from django.contrib import admin

from .views import report_csp

admin.autodiscover()


urlpatterns = [
    url(r'^report/$', report_csp, name='report_csp'),
]



# STANDARD LIB
from __future__ import unicode_literals

import json

from django.db import connection, models
from django.utils.html import escape
from django.utils.safestring import mark_safe

DISPOSITIONS = (
    ('enforce', 'enforce'),
    ('report', 'report'),
)

# Map of required CSP report fields to model fields
REQUIRED_FIELD_MAP = (
    ('document-uri', 'document_uri'),
    ('referrer', 'referrer'),
    ('blocked-uri', 'blocked_uri'),
    ('violated-directive', 'violated_directive'),
    ('original-policy', 'original_policy'),
)
# Map of optional (CSP >= 2.0) CSP report fields to model fields
OPTIONAL_FIELD_MAP = (
    ('effective-directive', 'effective_directive'),
    ('source-file', 'source_file'),
)
# Map of optional report fields with integer values.
INTEGER_FIELD_MAP = (
    ('status-code', 'status_code'),
    ('line-number', 'line_number'),
    ('column-number', 'column_number'),
)


class CSPReport(models.Model):
    """Represents a CSP violation report.

    @ivar created: Date and time of report creation.
    @ivar modified: Date and time of last modification.
    @ivar json: Raw CSP report
    @ivar is_valid: Whether the CSP report is valid.

    All report fields are NULL is report is invalid.
    Report fields are NULL if the report follows the rules of lower protocol level.

    CSP 1.0 fields:
    @ivar document_uri: The URI of protected resource.
    @ivar referrer: The referrer attribute of the protected resource.
    @ivar blocked_uri: The URI of the resource that was prevented from loading due to the policy violation.
    @ivar violated_directive: The policy directive that was violated.
    @ivar original_policy: The original policy as received by the user-agent.

    CSP 2.0 fields
    @ivar effective_directive: The name of the policy directive that was violated.
    @ivar status_code: The status-code of the HTTP response that contained the protected resource.
    @ivar source_file: The URL of the resource where the violation occurred.
    @ivar line_number: The line number in source-file on which the violation occurred.
    @ivar column_number: The column number in source-file on which the violation occurred.

    CSP 3.0 fields
    @ivar disposition: The disposition of violation's policy.
    """

    class Meta(object):
        ordering = ('-created',)

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    user_agent = models.TextField(blank=True)
    json = models.TextField()
    is_valid = models.BooleanField(default=False)
    # Individual report fields - use `TextField` because there are no limits by any specification for these fields.
    document_uri = models.TextField(blank=True, null=True)
    referrer = models.TextField(blank=True, null=True)
    blocked_uri = models.TextField(blank=True, null=True)
    violated_directive = models.TextField(blank=True, null=True)
    original_policy = models.TextField(blank=True, null=True)
    effective_directive = models.TextField(blank=True, null=True)
    status_code = models.PositiveSmallIntegerField(blank=True, null=True)
    source_file = models.TextField(blank=True, null=True)
    line_number = models.PositiveIntegerField(blank=True, null=True)
    column_number = models.PositiveIntegerField(blank=True, null=True)
    disposition = models.CharField(max_length=10, blank=True, null=True, choices=DISPOSITIONS)

    @property
    def nice_report(self):
        """Return a nicely formatted original report."""
        if not self.json:
            return '[no CSP report data]'
        try:
            data = json.loads(self.json)
        except ValueError:
            return "Invalid CSP report: '{}'".format(self.json)
        if 'csp-report' not in data:
            return 'Invalid CSP report: ' + json.dumps(data, indent=4, sort_keys=True, separators=(',', ': '))
        return json.dumps(data['csp-report'], indent=4, sort_keys=True, separators=(',', ': '))

    def __str__(self):
        return self.nice_report

    @classmethod
    def from_message(cls, message):
        """Creates an instance from CSP report message.

        If the message is not valid, the result will still have as much fields set as possible.

        @param message: JSON encoded CSP report.
        @type message: text
        """
        self = cls(json=message)
        try:
            decoded_data = json.loads(message)
        except ValueError:
            # Message is not a valid JSON. Return as invalid.
            return self
        try:
            report_data = decoded_data['csp-report']
        except KeyError:
            # Message is not a valid CSP report. Return as invalid.
            return self

        # Extract individual fields
        for report_name, field_name in REQUIRED_FIELD_MAP + OPTIONAL_FIELD_MAP:
            setattr(self, field_name, report_data.get(report_name))
        # Extract integer fields
        for report_name, field_name in INTEGER_FIELD_MAP:
            value = report_data.get(report_name)
            field = self._meta.get_field(field_name)
            min_value, max_value = connection.ops.integer_field_range(field.get_internal_type())
            if min_value is None:
                min_value = 0
            # All these fields are possitive. Value can't be negative.
            min_value = max(min_value, 0)
            if value is not None and min_value <= value and (max_value is None or value <= max_value):
                setattr(self, field_name, value)
        # Extract disposition
        disposition = report_data.get('disposition')
        if disposition in dict(DISPOSITIONS).keys():
            self.disposition = disposition

        # Check if report is valid
        is_valid = True
        for field_name in dict(REQUIRED_FIELD_MAP).values():
            if getattr(self, field_name) is None:
                is_valid = False
                break
        self.is_valid = is_valid

        return self

    @property
    def data(self):
        """ Returns self.json loaded as a python object. """
        try:
            data = self._data
        except AttributeError:
            data = self._data = json.loads(self.json)
        return data

    def json_as_html(self):
        """ Print out self.json in a nice way. """

        # To avoid circular import
        from cspreports import utils

        formatted_json = utils.format_report(self.json)
        return mark_safe("<pre>\n%s</pre>" % escape(formatted_json))



#!/usr/bin/env python
from setuptools import find_packages, setup

PACKAGES = find_packages()
REQUIREMENTS = ['django >=1.8,<2.3', 'six']
TEST_REQUIREMENTS = ['mock', 'coverage']
EXTRAS_REQUIRE = {
    'quality': ['isort', 'flake8'],
    'test': TEST_REQUIREMENTS,
}
CLASSIFIERS = ['License :: OSI Approved :: MIT License',
               'Framework :: Django',
               'Programming Language :: Python',
               'Programming Language :: Python :: 2',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3',
               'Programming Language :: Python :: 3.4',
               'Programming Language :: Python :: 3.5',
               'Programming Language :: Python :: 3.6']


DESCRIPTION = (
    "A Django app for handling reports from web browsers of violations of your website's "
    "HTTP Content Security Policy."
)

setup(
    name='django-csp-reports',
    version='1.3',
    description=DESCRIPTION,
    long_description=DESCRIPTION,
    author='Adam Alton',
    author_email='adamalton@gmail.com',
    url='https://github.com/adamalton/django-csp-reports',
    download_url='https://github.com/adamalton/django-csp-reports/tarball/1.3',
    packages=PACKAGES,
    include_package_data=True,
    python_requires='>=2.7',
    install_requires=REQUIREMENTS,
    tests_require=TEST_REQUIREMENTS,
    extras_require=EXTRAS_REQUIRE,
    test_suite='runtests.runtests',
    keywords=['django', 'csp', 'content security policy'],
    classifiers=CLASSIFIERS,
)



"""Test commands."""
from __future__ import unicode_literals

from datetime import datetime

from django.core.management import CommandError, call_command
from django.test import TestCase, override_settings
from django.utils import timezone
from mock import patch
from six import StringIO

from cspreports.models import CSPReport

from .utils import create_csp_report


class TestCleanCspreports(TestCase):
    """Test `clean_cspreports` command."""

    def create_cspreport(self, created):
        report = CSPReport.objects.create()
        report.created = created
        report.save()
        return report

    def test_empty(self):
        # Test case when there's nothing to clean
        call_command('clean_cspreports')

    def test_verbose(self):
        # Test verbose prints message
        buff = StringIO()
        call_command('clean_cspreports', verbosity=2, stdout=buff)
        self.assertIn('Deleted all reports ', buff.getvalue())

    @override_settings(USE_TZ=True, TIME_ZONE='Europe/Prague')
    def test_clean(self):
        # Test reports are cleaned correctly
        self.create_cspreport(datetime(2016, 4, 19, 21, 59, 59, tzinfo=timezone.utc))
        self.create_cspreport(datetime(2016, 4, 19, 22, 0, 0, tzinfo=timezone.utc))
        mock_now = datetime(2016, 4, 27, 0, 34, tzinfo=timezone.utc)
        with patch('cspreports.utils.now', return_value=mock_now):
            call_command('clean_cspreports')

        self.assertQuerysetEqual(CSPReport.objects.values_list('created'),
                                 [(datetime(2016, 4, 19, 22, 0, 0, tzinfo=timezone.utc), )],
                                 transform=tuple)

    def test_invalid(self):
        # Test invalid limit input
        with self.assertRaisesMessage(CommandError, "'JUNK' is not a valid date."):
            call_command('clean_cspreports', 'JUNK')


class TestMakeCspSummary(TestCase):
    """Test `make_csp_summary` command."""

    def test_empty(self):
        buff = StringIO()
        call_command('make_csp_summary', stdout=buff)
        self.assertIn('CSP report summary', buff.getvalue())

    def test_report(self):
        create_csp_report(datetime(1970, 1, 1, 12), is_valid=True, document_uri='http://example.cz/',
                          blocked_uri='http://example.evil/')
        buff = StringIO()

        call_command('make_csp_summary', stdout=buff, since='1970-01-01')

        self.assertIn('CSP report summary', buff.getvalue())
        self.assertIn('Violation sources', buff.getvalue())
        self.assertIn('Blocked URIs', buff.getvalue())
        self.assertIn('http://example.evil/', buff.getvalue())

    def test_invalid_since(self):
        # Test invalid since value
        with self.assertRaisesMessage(CommandError, "'JUNK' is not a valid date."):
            call_command('make_csp_summary', since='JUNK')

    def test_invalid_to(self):
        # Test invalid since to
        with self.assertRaisesMessage(CommandError, "'JUNK' is not a valid date."):
            call_command('make_csp_summary', to='JUNK')



"""Test utilities."""
from __future__ import unicode_literals

from cspreports.models import CSPReport


def create_csp_report(created, **kwargs):
    """Create and return a CSP report with a custom created timestamp."""
    # Since the `created` is filled on save, we have to override it manually.
    report = CSPReport.objects.create(**kwargs)
    report.created = created
    report.save()
    return report
